import numpy as np
import matplotlib.pyplot as plt
import numpy.linalg as la


def berry_curvature(H, Nx, Ny=None, xfrac=1,
                      plot=True, fname=None, nocc=1, periodic=True, plot_kwargs=dict(), r=None):
    """
    Berry curvature calculation using projectors

    Parameters:
    -----------

    H: callable
        Hamiltonian H(kx, ky).
    Nx, Ny: int
        Number of smapling points in k-space, using Nx x Ny grid.
    xfrac: int (optional)
        If magnetic translation symmetry is present, only sample
        1/xfrac part of the BZ in the k_x direction. For KM model
        xfrac = q.
    nocc: int (optional)
        Number of occupied bands of lowest energy.
    periodic: bool (optional)
        Whether the Hamiltonian is in the BZ periodic Bloch basis.
        The "periodic gauge" corresponds to False.
    r: 2D array (optional)
        Array of site coordinates. If the Hamiltonian is in the
        "periodic gauge", no need to provide it. If it is in the
        BZ periodic gauge, provide r to get the correct Berry
        curvature corresponding to EM responses.
    plot: bool (optional)
        Whether to plot the curvature.
    fname: string (optional)
        If provided, the figure is saved in this file.
    plot_kwargs: dict (optional)
        Parameters to pass on to plt.imshow


    Returns:
    --------

    eta: 2D array
        Scalar field describing the Berry curvature F(k).
    """
    eta, kxs, kys = quantum_metric(H, Nx, Ny, xfrac, nocc, periodic, r, return_ks=True)
    bc = -(eta[0, 1] - eta[1, 0]).imag

    if plot:
        plot_curvature(kxs, kys, bc, xfrac, fname, **plot_kwargs)
        
    return bc


def quantum_metric(H, Nx, Ny=None, xfrac=1, nocc=1, periodic=True, r=None, return_ks=False):
    """
    Quantum metric calculation using projectors

    Parameters:
    -----------

    H: callable
        Hamiltonian H(kx, ky).
    Nx, Ny: int
        Number of smapling points in k-space, using Nx x Ny grid.
    xfrac: int (optional)
        If magnetic translation symmetry is present, only sample
        1/xfrac part of the BZ in the k_x direction. For KM model
        xfrac = q.
    nocc: int (optional)
        Number of occupied bands of lowest energy.
    periodic: bool (optional)
        Whether the Hamiltonian is in the BZ periodic Bloch basis.
        The "periodic gauge" corresponds to False.
    r: 2D array (optional)
        Array of site coordinates. If the Hamiltonian is in the
        "periodic gauge", no need to provide it. If it is in the
        BZ periodic gauge, provide r to get the correct Berry
        curvature corresponding to EM responses.
    return_ks: (optional)
        Whether to return the list of k-points.

    Returns:
    --------

    eta: 4D array
        Tensor field describing the quantum metric eta(k),
        the last two axes correspond to the tensor components.

    eta, kxs, kys: 4D array, 1D array, 1D array
        If return_ks is True, also return the sampling points
        in k-space.
    """
    Ny = Nx if Ny is None else Ny
    klim = np.pi if periodic else np.pi * (1 + 2/Nx)
    kxs, dkx = np.linspace(-klim / xfrac, klim / xfrac,
                          (Nx if periodic else Nx + 2), endpoint=False, retstep=True)
    klim = np.pi if periodic else np.pi * (1 + 2/Ny)
    kys, dky = np.linspace(-klim, klim,
                          (Ny if periodic else Ny + 2), endpoint=False, retstep=True)

    # calculation of projectors on the grid
    ham = np.array([[H(kx, ky) for ky in kys] for kx in kxs])

    # Transform to basis with real space Bloch factors if r is provided
    if r is not None:
        W = np.array([[np.diag(np.exp(1j * np.dot(r, [kx, ky]))) for ky in kys] for kx in kxs])
        ham = np.transpose(W.conj(), (0, 1, 3, 2)) @ ham @ W

    eigvals, eigvecs = la.eigh(ham)
    eigvecs = eigvecs[:, :, :, :nocc]
    P = eigvecs @ np.transpose(eigvecs.conj(), (0, 1, 3, 2))

    # berry curvature calculation
    DP = grad(P, axes=(0, 1), steps=(dkx, dky))

    if not periodic:
        DP, P = DP[:, 1: -1, 1: -1], P[1: -1, 1: -1]
        kxs = kxs[1: -1]
        kys = kys[1: -1]

    eta = np.trace(DP[:, np.newaxis] @ DP[np.newaxis, :] @ P[np.newaxis, np.newaxis], axis1=-1, axis2=-2)

    if return_ks:
        return eta, kxs, kys
    else:
        return eta


def plot_curvature(kx, ky, z, xfrac=1, fname=None, figsize=(3, 2.5), **kwargs):
    norm = np.mean(z)
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_xlabel(r'$k_x$')
    ax.set_ylabel(r'$k_y$')
    c = ax.imshow(z/norm,
                  aspect=1,
                  extent=[min(kx), max(kx), min(ky)/xfrac, max(ky)/xfrac],
                  **kwargs)
    ax.set_xticks([min(kx), 0, max(kx)])
    ax.set_yticks([min(ky)/xfrac, 0, max(ky)/xfrac])
    if xfrac == 1:
        ticks = [fr'$-\pi$', r'$0$', fr'$\pi$']
    else:
        ticks = [fr'$-\frac{{\pi}}{{{xfrac}}}$', r'$0$', fr'$\frac{{\pi}}{{{xfrac}}}$']
    ax.set_xticklabels(ticks)
    ax.set_yticklabels(ticks)
    plt.colorbar(c)
    # Get the current axis 
    ax = plt.gca()        
    # Get the images on an axis
    im = ax.images        
    # Assume colorbar was plotted last one plotted last
    cb = im[-1].colorbar
    # Force scientific notation
    cb.formatter.set_scientific(True)
    cb.formatter.set_powerlimits((-1,1))

    if fname is not None:
        plt.savefig(fname, bbox_inches='tight', dpi=400)
    plt.show()


def k_transform(kx, ky, A, N, q):
    k = np.array([kx, ky/q])
    k_ind = np.array(np.around(q * k * (N / (2 * np.pi))), dtype=int)
    k_new = k - A[k_ind[1] % N, k_ind[0] % N, :] / q
    return k_new * np.array([1, q])


def flatten_curvature(curvature):
    """Perform a single step of the flattening algorythm.

    Parameters:
    -----------

    curvature: 2D array
        Curvature function F(k) to be flattened, sampled on a
        discrete k-grid.

    Returns:
    --------

    A: 3D array
        Vector field describing the displacement field h(k),
        the last axis corresponds to the two components of h.
    """
    B = np.fft.ifft2(np.fft.fftshift(curvature) / np.mean(curvature))

    N = curvature.shape[0]
    s = slice(0, N)
    neighbors = np.mgrid[s, s].T - np.array([1, 1]) * N//2
    denom = np.einsum('xyi, xyi -> xy', neighbors, N/(2 * np.pi) * np.sin((2 * np.pi) / N * neighbors))
    A = np.fft.fftshift(B)[:, :, np.newaxis] * neighbors / denom[:, :, np.newaxis]
    A = np.nan_to_num(A)
    A[0, :, 1] = A[:, 0, 0] = 0
    A = np.fft.fftshift(A, axes=(0, 1))
    A = (1j * np.fft.fft2(A, axes=(0, 1))).real

    return A


def iterate_flatten_curvature(H, N, iterations=10, A0=None, xfrac=1, periodic=False,
                              plot=True, verbose=True, stop_increase=False):
    """Perform the flattening algorythm.

    Parameters:
    -----------

    H: callable
        Hamiltonian H(kx, ky).
    N: int
        Number of smapling points in k-space, using N x N grid.
    iterations: int
        Maximum number of iterations.
    A0: 3D array (optional)
        Starting displacement field h(k).
    xfrac: int (optional)
        If magnetic translation symmetry is present, only sample
        1/xfrac part of the BZ in the k_x direction. For KM model
        xfrac = q.
    periodic: bool (optional)
        Whether the Hamiltonian is in the BZ periodic Bloch basis.
        The "periodic gauge" corresponds to False.
    plot: bool (optional)
        Whether to plot the curvature in everyt step.
    verbose: bool (optional)
        Whether to give detailed output.
    stop_increase: bool (optional)
        Whether to stop the iteration if the variance of the
        curvature starts increasing. The result with smallest
        variance is returned.

    Returns:
    --------

    A: 3D array
        Vector field describing the displacement field h(k),
        the last axis corresponds to the two components of h.
    """
    A = np.zeros((N, N, 2)) if A0 is None else A0
    H_new = H
    stdF = np.inf
    if plot:
        F = berry_curvature(H_new, N, plot=plot, periodic=periodic, xfrac=xfrac)
    if verbose:
        print('iteration', 0, ', C = ', np.sum(F) * ((2 * np.pi) / F.shape[0])**2 / (2*np.pi), ', std(F) =', np.std(F))

    for i in range(iterations):
        A_new = flatten_curvature(F)
        def H_new(kx, ky):
            return H(*k_transform(kx, ky, A + A_new, N, xfrac))
        F = berry_curvature(H_new, N, plot=plot, periodic=periodic, xfrac=xfrac)
        if verbose:
            print('iteration', i + 1, ', C = ', np.sum(F) * ((2 * np.pi) / F.shape[0])**2 / (2*np.pi), ', std(F) =', np.std(F))
        if stdF < np.std(F) and stop_increase:
            # if the curvature variation increased, return the previous Hamiltonian
            def H_new(kx, ky):
                return H(*k_transform(kx, ky, A, N, xfrac))
            break
        stdF = np.std(F)
        A += A_new

    return A, H_new


# Various geometric functions
def grad(A, axes=(0, 1), steps=None, ranges=None):
    if steps is None:
        ranges = np.ones((len(axes),)) if ranges is None else ranges
        steps = ranges / np.array(A.shape)[list(axes)]
    return np.array([
            (np.roll(A, -1, a) - np.roll(A, 1, a)) / (2 * dxa)
            for a, dxa in zip(axes, steps)
    ])


def g_inv(g):
    return np.moveaxis(la.inv(np.moveaxis(g, (0, 1), (2, 3))), (0, 1), (2, 3))


def christoffel(g, gi=None):
    dg = grad(g, axes=(-1, -2))
    gi = g_inv(g) if gi is None else gi
    return 1/2 * (np.einsum('imxy,lmkxy->iklxy', gi, dg) +
                  np.einsum('imxy,kmlxy->iklxy', gi, dg) -
                  np.einsum('imxy,mklxy->iklxy', gi, dg))


def ricci(gam):
    dgam = grad(gam, axes=(-1, -2))
    return (np.einsum('iijkxy->jkxy', dgam) -
            np.einsum('jiikxy->jkxy', dgam) +
            np.einsum('iipxy,pjkxy->jkxy', gam, gam) -
            np.einsum('ijpxy,pikxy->jkxy', gam, gam))


def scalar_curvature(g):
    gi = g_inv(g)
    gam = christoffel(g, gi)
    R = ricci(gam)
    return np.einsum('ijxy,jixy->xy', gi, R)
