import numpy as np
from collections import defaultdict
import kwant
import tinyarray as ta


def kapit_mueller(q, ran=3, finalized=True, plot=False):
    """Kapit-Mueller Hamiltonian in kwant format.

    q: int
        model with 1/q flux quantum per unit cell
    ran: int
        hopping range after which hoppings are truncated
    J0: dict (optional)
        optionally modified hopping prefactors
    """
    syst = kwant.Builder(symmetry=kwant.lattice.TranslationalSymmetry([1, 0], [0, q]))
    lat = kwant.lattice.square()

    def hopping(site1, site2, J0):
        x, y = site1.pos - site2.pos
        z = x + 1j * y
        xj, yj = site1.pos
        zj = xj + 1j * yj
        xk, yk = site2.pos
        zk = xk + 1j * yk
        W = (-1)**(1+x+y+x*y) * np.exp(-(np.pi/2) * (1-1/q) * np.abs(z)**2)
        # Landau gauge
        gauge = np.exp(1j * np.pi / q * (xj * yj - xk * yk))
        hop = J0[x, y] * gauge * W * np.exp(np.pi/2 * (zj * np.conj(z) - np.conj(zj) * z) * 1/q)
        # print(x, y, hop)
        return hop

    for i in range(q):
        syst[lat(0, i)] = 0

    for n in range(1, ran + 1):
        syst[lat.neighbors(n)] = hopping

    wrapped = kwant.wraparound.wraparound(syst)

    if plot:
        kwant.plot(syst);

    if finalized:
        return wrapped.finalized()
    else:
        return wrapped


def H_KM(q, ran=3, J0=None):
    """Kapit-Mueller Hamiltonian.

    q: int
        model with 1/q flux quantum per unit cell
    ran: int
        hopping range after which hoppings are truncated
    J0: dict (optional)
        optionally modified hopping prefactors
    """

    if J0 is None:
        J0 = defaultdict(lambda : -1)

    def hopping(site1, site2, J0):
        x, y = site1 - site2
        z = x + 1j * y
        xj, yj = site1
        zj = xj + 1j * yj
        xk, yk = site2
        zk = xk + 1j * yk
        W = (-1.)**(1+x+y+x*y) * np.exp(-(np.pi/2) * (1-1/q) * np.abs(z)**2)
        # Landau gauge
        gauge = np.exp(1j * np.pi / q * (xj * yj - xk * yk))
        hop = J0[x, y] * gauge * W * np.exp(np.pi/2 * (zj * np.conj(z) - np.conj(zj) * z) * 1/q)
        # print(x, y, hop)
        return hop

    H_dict = defaultdict(lambda : np.zeros((q, q), dtype=complex))
    # this is the only place kwant is used in this function
    neighbors = kwant.linalg.lll.cvp((0, 0), np.eye(2), n=ran+1, group_by_length=True)

    for i in range(q):
        site1 = np.array([0, i])
        for d in neighbors:
            site2 = site1 + d
            j = ta.array(site2[1] % q)
            H_dict[ta.array(d)][i, j] = hopping(site1, site2, J0)

    return H_dict, H_from_dict(H_dict)


def H_from_dict(H_dict):
    q = H_dict[0, 0].shape[0]
    def H(kx, ky):
        return np.sum([h * np.exp(1j * np.dot([kx, ky/q], d)) for d, h in H_dict.items()],
                      axis=0)
    return H


def C3model_kwant(finalized=True, plot=False):
    """Chern number 3 model from PhysRevB.96.165150"""
    hNN = np.array([[6.65, 4j, -6],
                    [4j, -4.87805, 7.57576j],
                    [-6, 7.57576j, -1.78571]])
    hNNN = np.array([[-1.10833, -3.125 * (1 + 1j), -3.5j],
                     [3.125 * (1 - 1j), -2.94118, 1.66667 * (1 + 1j)],
                     [3.5j, -1.66667 * (1 - 1j), 4]])


    syst = kwant.Builder(symmetry=kwant.lattice.TranslationalSymmetry([1, 0], [0, 1]))
    lat = kwant.lattice.square(norbs=3)

    syst[lat(0, 0)] = np.zeros((3, 3))

    syst[lat(0, 0), lat(1, 0)] = hNN
    syst[lat(0, 0), lat(1, 1)] = hNNN

    C4 = np.diag([1, 1j, -1])
    assert np.allclose(C4 @ C4 @ hNN @ C4.T.conj() @ C4.T.conj(), hNN.T.conj())
    assert np.allclose(C4 @ C4 @ hNNN @ C4.T.conj() @ C4.T.conj(), hNNN.T.conj())
    syst[lat(0, 0), lat(0, 1)] = C4 @ hNN @ C4.T.conj()
    syst[lat(0, 0), lat(-1, 1)] = C4 @ hNNN @ C4.T.conj()

    wrapped = kwant.wraparound.wraparound(syst)

    if plot:
        kwant.plot(syst);

    if finalized:
        return wrapped.finalized()
    else:
        return wrapped


def C3model():
    """Chern number 3 model from PhysRevB.96.165150"""
    hNN = np.array([[6.65, 4j, -6],
                    [4j, -4.87805, 7.57576j],
                    [-6, 7.57576j, -1.78571]])
    hNNN = np.array([[-1.10833, -3.125 * (1 + 1j), -3.5j],
                     [3.125 * (1 - 1j), -2.94118, 1.66667 * (1 + 1j)],
                     [3.5j, -1.66667 * (1 - 1j), 4]])

    C4 = np.array(np.diag([1, 1j, -1]), dtype=complex)
    assert np.allclose(C4 @ C4 @ hNN @ C4.T.conj() @ C4.T.conj(), hNN.T.conj())
    assert np.allclose(C4 @ C4 @ hNNN @ C4.T.conj() @ C4.T.conj(), hNNN.T.conj())
    hNNr = C4 @ hNN @ C4.T.conj()
    hNNNr = C4 @ hNNN @ C4.T.conj()

    def H(kx, ky):
        return np.sum([h * np.exp(1j * np.dot([kx, ky], d))
                       + h.T.conj() * np.exp(-1j * np.dot([kx, ky], d))
            for h, d in zip([hNN, hNNr, hNNN, hNNNr],
                            [[1, 0], [0, 1], [1, 1], [-1, 1]])], axis=0)

    return H
